function [s, fcnEvals, lineSearchSuccess] = strongWolfe(fcn, x, s, t, f, g, fcnEvals, param)
%
% function [s, fcnEvals, lineSearchSuccess] = strongWolfe(fcn, x, s, t, f, g, fcnEvals, param)
%
% Authors:
%   (c) Matthias Chung (e-mail: mcchung@vt.edu) in April 2014
%
% MATLAB Version: 8.2.0.701 (R2013b)
%
% Description:
%   The strong Wolfe line search algorithm using Hermite interpolation
%
% Input arguments:
%   fcn               - objective function
%   x                 - argument value
%   s                 - search direction
%   f                 - recent function value
%   g                 - recent gradient value
%   t                 - initial step length parameter t < tMax
%   fcnEvals          - number of function evaluations
%   #param            - further options of algorithm
%     c1              - Armijo condition number [ 1e-4 ]
%     c2              - curvature condtion number [ 0.9 ]
%     tMax            - maximal step length factor [ 10 ]
%     maxIter         - maximum number of iterations [ 10 ]
%     sgtol           - interpolation safe guart value [ 1e-5 ]
%
% Output arguments:
%   s                 - update vector
%   fcnEvals          - number of function evaluations
%   lineSearchSuccess - returns flag of line search success 0 or 1
%
% Details:
%
% References:
%   [1] Nocedal, J. and Wright, S. J., Numerical Optimization,
%       Springer, New York, 2006.
%

% initialize default options of algorithm
c1 = 1e-4; c2 = 0.9; tMax = 10; maxIter = 10; sgtol = 1e-5;

% rewrite default parameters if needed
if nargin == nargin(mfilename)
  for j = 1:size(param,1), eval([param{j,1},'= param{j,2};']); end
end

df0 = g'*s; % calculate s directional derivative at x 

% check if s is descent direction
if df0 > 0
  lineSearchSuccess = 0; 
    warning('Matlab:strongWolfe',...
    'No descent direction selected for line search.');
else % perform strong Wolfe line search
  [s, fcnEvals, lineSearchSuccess] = strongWolfeStep(fcn, x, s, f, df0, c1, c2, fcnEvals, t, tMax, maxIter,sgtol);
end

if ~lineSearchSuccess
  warning('Matlab:strongWolfe',...
    'Strong Wolfe line search does not succeed. Return with recent values.');
end

end

% ----------------------------------------------------------------------------------------------------------------------------

function [s, fcnEvals, lineSearchSuccess] = strongWolfeStep(fcn, x, s, f0, df0, c1, c2, fcnEvals, t, tMax, maxIter,sgtol)
% strong Wolfe line search step

% initialize fixed parameters of algorithm
t(2) = t; t(1) = 0; f(1) = f0; lineSearchSuccess = 1; iter = 0;

% pre-compute directional derivative times scalar c1 and c2
df(1) = df0;
c1df0  =  c1*df0; % compute directional derivative at f0 with factor c1
mc2df0 = -c2*df0; % compute negative directional derivative at f0 with factor c2

while iter < maxIter % expand to find interval of strong Wolfe points
  
  iter = iter+1; % track number of iterations
  
  [f(2), g] = fcn(x + t(2)*s); df(2) = g'*s; % compute function values at new point
  
  if f(2) > f0 + t(2)*c1df0 || f(2) >= f(1) % case one: Armijo condition not fulfilled, new point ensured to be to the right of the Wolfe region
    % above start zooming  (Hermite interpolation condition not fulfilled.)
    [tM, fcnEvals, lineSearchSuccess] = zoom(fcn,x,s,f0,t,f,df,c1df0,mc2df0,fcnEvals,maxIter,sgtol);
    break
  end
  
  if abs(df(2)) <= mc2df0,
    tM = t(2);
    break
  end % case two: both curvature and Armijo conditions fulfilled, line search successful, stop line search
  
  if 	df(2) >= 0  % case three: Armijo condition fulfilled, and positive curvature at right interval end is ensured. Hence df(1) < 0 and df(2) > 0 we can use Hermite interpolation
    [tM, fcnEvals, lineSearchSuccess] = zoom(fcn,x,s,f0,t,f,df,c1df0,mc2df0,fcnEvals,maxIter,sgtol);
    break
  end
  
  t(2) = 0.5*(t(1)+tMax);  % expansion step (relative expansion)
  
end

fcnEvals(1:2) = fcnEvals(1:2)+iter;

if ~exist('tM','var'), tM = 0; end
if tM <= eps || iter == maxIter, lineSearchSuccess = 0; end	% failure of strong Wolfe line search

s = tM*s;

end

% ----------------------------------------------------------------------------------------------------------------------------

function [tM, fcnEvals, lineSearchSuccess] = zoom(fcn,x,s,f0,t,f,df,c1f0,mc2df0,fcnEvals,maxIter,sgtol)

lineSearchSuccess = 1;

iter = 0;
while iter < maxIter
  iter = iter + 1;

  tM = findMinimizer(t,f,df,sgtol); % get Hermite polynomial minimizer

  [fM, g] = fcn(x + tM*s); dfM = g'*s;
  
  if fM > f0 + tM*c1f0 || fM >= f(1)
    t(2) = tM; f(2) = fM; df(2) = dfM; % update t2
  else
    
    if abs(dfM) <= mc2df0,
      break
    end
    if (dfM)*(t(2)-t(1)) >= 0 % update right interval point t(2)
      t(2) = t(1);
      f(2) = f(1);
      df(2) = df(1);
    end
    
    t(1) = tM; f(1) = fM; df(1) = dfM; % update left interval point
    
    if t(1) > t(2)
      t = flipud(t); f = flipud(f); df = flipud(df);
    end
  end
  
  fcnEvals(1:2) = fcnEvals(1:2)+iter;
  
  if iter == maxIter, lineSearchSuccess = 0; end
  
end

end

% ----------------------------------------------------------------------------------------------------------------------------

function tM = findMinimizer(t,f,df,sgtol) % using Hermite Polynomial minimization
% inputs: t  - intervall [t(1), t(2)] with t(2) > t(1)
%         f  - function values at t
%         df - derivative of f at t
% returns:
%         tM - minimizer of Hermite polynomial in interval [t(1), t(2)]

% ensure df(1) < 0 and df(2) > 0
if df(1) >= 0 || df(2) <= 0
  tM = 0.5*sum(t); % use mid-point
else % cubic Hermite interpolation
  
  % compute minimum of Hermite polynomial in interval [t(1), t(2)]
  z = 3*(f(2)-f(1))/(t(2)-t(1));
  z = z-df(2)-df(1); % re-use of z
  tM = sqrt(z^2-df(2)*df(1));
  
  % minimizer in interval given by
  tM = t(1)+(t(2)-t(1))*(tM-df(1)-z)/(df(2)-df(1)+2*tM); % re-use of z
  
  % safe guard not to close to interval ends
  if abs(tM - t(1)) < sgtol
    tM = t(1) + sgtol;
  elseif abs(t(2) - tM) < sqrt(eps)
    tM = t(2) - sgtol;
  end
  
end

end